/* Configuration for an i386 running emx (target). */

/* Use #ifdef __EMX__ for issues needed by EMX host */
/* Use #ifdef EMX for issues needed for a EMX target */
/* Note that the cross-compiler is broken, it may need adjustments */
#define EMX

/* We do want to add an underscore to the front of each user symbol. */
#define YES_UNDERSCORES

#include "i386/gstabs.h" /* GAS stabs */
#include "aoutos.h"	 /* constructors/destructors/garbage collectors */

/* The system headers are C++-aware. */
#define NO_IMPLICIT_EXTERN_C

/* Don't provide default values for __CTOR_LIST__ and __DTOR_LIST__ in libgcc */
#define CTOR_LISTS_DEFINED_EXTERNALLY

/* Enable #pragma pack. */
#define HANDLE_SYSV_PRAGMA

/* Handle drive separator in path names */
#define HAVE_DOS_BASED_FILESYSTEM
#define HAVE_DOS_BASED_FILE_SYSTEM

/* Don't automatically force .exe suffix */
#define NO_FORCE_EXEOBJ_SUFFIX

/* Use `#' instead of `/' as assembler comments */
#undef ASM_COMMENT_START
#define ASM_COMMENT_START "#"

/* Support for C++ templates.  */
#undef MAKE_DECL_ONE_ONLY
#define MAKE_DECL_ONE_ONLY(DECL) (DECL_WEAK (DECL) = 1)

/* This is how we tell the assembler that a symbol is weak.  */
#define ASM_WEAKEN_LABEL(FILE, NAME) 	\
  do					\
    {					\
      fputs ("\t.weak\t", (FILE));	\
      assemble_name ((FILE), (NAME)); 	\
      fputc ('\n', (FILE));		\
    }					\
  while (0)

/* We want DWARF2 frame unwind info for exception handling */
#define DWARF2_UNWIND_INFO 1

/* This macros will stick a label to exception table for current file,
 * and also will put it in a 'ctor/dtor' - like list (using .stabs - this uses
 * same technique used in aoutos.h) which ld will collect into a single table
 * and then __ctordtorInit will use this table to make a call to
 * __register_frame_table
 */
#define EH_FRAME_SECTION()						       \
  data_section ();							       \
  fprintf (asm_out_file, "\t.stabs\t\"___eh_frame__\",24,0,0,Lframe%d\n", for_eh);

/* Put exception tables into .data section */
#define EXCEPTION_SECTION()						       \
  data_section()

/* Terminate DWARF frame unwind info with a closing zero */
#define DWARF2_FRAMEINFO_NULL_TERMINATED 1

/* Generate stack probes for allocations bigger than */
#define CHECK_STACK_LIMIT 0x1000

/* wchar_t is unsigned short. */
#define WCHAR_TYPE              "short unsigned int"
#define WCHAR_TYPE_SIZE         16

/* If bit field type is int, don't let it cross an int,
   and give entire struct the alignment of an int.  */
/* Required on the 386 since it doesn't have bitfield insns.  */
#define PCC_BITFIELD_TYPE_MATTERS 1

/* The svr4 ABI for the i386 says that records and unions are returned
   in memory.  */
#undef DEFAULT_PCC_STRUCT_RETURN
#define DEFAULT_PCC_STRUCT_RETURN 1

/* Don't continue symbols.  I think all relevant tools can handle
   symbols of arbitrary length.  GAS does not correctly handle
   continued symbols: it sometimes emits another symbol before the
   continuation. */
#define DBX_CONTIN_LENGTH 0

/* A C statement (sans semicolon) to output to the stdio stream
   FILE the assembler definition of uninitialized global DECL named
   NAME whose size is SIZE bytes and alignment is ALIGN bytes.
   Try to use asm_output_aligned_bss to implement this macro.  */
#define ASM_OUTPUT_ALIGNED_BSS(FILE, DECL, NAME, SIZE, ALIGN) \
  asm_output_aligned_bss (FILE, DECL, NAME, SIZE, ALIGN)

/******************************************************************************
 *                       - S - W - I - T - C - H - E - S -
 ******************************************************************************/

/* The -Zstack option takes an argument. */
#define WORD_SWITCH_TAKES_ARG(STR) (DEFAULT_WORD_SWITCH_TAKES_ARG (STR)	       \
                                    || !strcmp (STR, "Zstack"))

/* Define an additional switch -m(no-)probe for backward compatibility */
/* Also define the -m(no-)epilogue switch */

#define MASK_EPILOGUE		0x40000000      /* Use FUNCTION_EPILOGUE */

#undef SUBTARGET_SWITCHES
#define SUBTARGET_SWITCHES						       \
    { "probe", MASK_STACK_PROBE, N_("Do stack probing")},		       \
    { "no-probe", -MASK_STACK_PROBE, N_("Don't do stack probing")},	       \
    { "epilogue", MASK_EPILOGUE, N_("Generate prologue/epilogue labels")},     \
    { "no-epilogue", -MASK_EPILOGUE, N_("Don't generate prologue/epilogue labels") },

/* Use FUNCTION_EPILOGUE */
#define TARGET_EPILOGUE (target_flags & MASK_EPILOGUE)

/* Add a __POST$xxx label before epilogue if -mepilogue specified */
#define FUNCTION_BEGIN_EPILOGUE(FILE)					       \
if (TARGET_EPILOGUE && TREE_PUBLIC (current_function_decl))		       \
{									       \
    char *func_label = IDENTIFIER_POINTER (DECL_ASSEMBLER_NAME (current_function_decl));\
    char *lbl = alloca (strlen (func_label) + 8);			       \
    strcpy (lbl, "__POST$");						       \
    strcat (lbl, func_label);						       \
    ASM_OUTPUT_LABEL (asm_out_file, lbl);				       \
}

/* Output assembler code to FILE to increment profiler label # LABELNO
   for profiling a function entry.  */

#undef FUNCTION_PROFILER
#define FUNCTION_PROFILER(FILE, LABELNO)				       \
{									       \
    fprintf (FILE, "\tcall __mcount\n");				       \
}

/* Calling conventions */

#undef VALID_MACHINE_TYPE_ATTRIBUTE
#define VALID_MACHINE_TYPE_ATTRIBUTE(TYPE, ATTRIBUTES, NAME, ARGS)	       \
    (emx_valid_type_attribute_p (TYPE, ATTRIBUTES, NAME, ARGS))

#undef RETURN_POPS_ARGS
#define RETURN_POPS_ARGS(FUNDECL,FUNTYPE,SIZE)				       \
    (emx_return_pops_args (FUNDECL, FUNTYPE, SIZE))

/******************************************************************************
 *                             - P - A - T - H - S -
 ******************************************************************************/

#define MD_STARTFILE_PREFIX STANDARD_EXEC_PREFIX

#define INCLUDE_DEFAULTS						       \
{									       \
  { GCC_INCLUDE_DIR, "GCC", 0, 1 },					       \
  { GCC_INCLUDE_DIR "/../../../../../include", "GCC", 0, 1 },		       \
  { GPLUSPLUS_INCLUDE_DIR, "G++", 1, 1 },				       \
  { GPLUSPLUS_TOOL_INCLUDE_DIR, "G++", 1, 1 },				       \
  { GPLUSPLUS_BACKWARD_INCLUDE_DIR, "G++", 1, 1 },			       \
  { 0, 0, 0, 0 }							       \
}

/* This macro is invoked during GCC startup to initialize exec_prefix etc */
#define INITIALIZE_PREFIXES						       \
{									       \
    char prefix_buffer [260 + 1];					       \
    if (!_execname (prefix_buffer, sizeof (prefix_buffer)))		       \
    {									       \
        char *name = _getname (prefix_buffer);				       \
        if (name > prefix_buffer)					       \
        {								       \
            *(--name) = 0;						       \
            name = _getname (prefix_buffer);				       \
        }								       \
        if (name > prefix_buffer)					       \
            name--;							       \
        strcpy (name, "/lib/gcc-lib/");					       \
        standard_exec_prefix =						       \
        standard_startfile_prefix = xstrdup (prefix_buffer);		       \
    }									       \
}

/* Dynamically compute the path prefix for GCC */
#ifdef STANDARD_STARTFILE_PREFIX
#define GET_ENV_PATH_LIST(VAR,NAME)					       \
{									       \
    (VAR) = getenv (NAME);						       \
    /* If we compare each char instead of doing strcmp() the following */      \
    /* block of code gets optimized out if it is not needed */		       \
    if (!(VAR))								       \
        if (NAME [0] == 'G' && NAME [1] == 'C' && NAME [2] == 'C'	       \
         && NAME [3] == '_' && NAME [4] == 'E' && NAME [5] == 'X'	       \
         && NAME [6] == 'E' && NAME [7] == 'C' && NAME [8] == '_'	       \
         && NAME [9] == 'P' && NAME [10] == 'R' && NAME [11] == 'E'	       \
         && NAME [12] == 'F' && NAME [13] == 'I' && NAME [14] == 'X'	       \
         && NAME [15] == 0)						       \
          (VAR) = (char *)standard_exec_prefix;					       \
}
#endif

/******************************************************************************
 *                             - S - P - E - C - S -
 ******************************************************************************/
/* Predefine symbols `__32BIT__', `__EMX__', and `i386'. */
#undef CPP_PREDEFINES
#define CPP_PREDEFINES "-D__32BIT__ -D__EMX__ -D__OS2__ "		       \
  "-Dunix -Asystem=unix -Asystem=posix -Asystem=emx"

/* Predefine the symbol `__MT__' for multi-threaded programs (-Zmt) */
#undef CPP_SPEC
#define CPP_SPEC \
  "%(cpp_cpu) "\
  "%{posix:-D_POSIX_SOURCE} "\
  "%{Zmt:-D__MT__} "\
  "%{pg:-D__GPROF__} "\
  "%{mepilogue:-D__EPILOGUE__} "\
  "%{mprobe:-D__STACK_PROBE__} "\
  "%{mstack-arg-probe:-D__STACK_PROBE__} "\
  "-D__cdecl=__attribute__((__cdecl__)) "\
  "-D__stdcall=__attribute__((__stdcall__)) "\
  "-D__optlink=__attribute__((__optlink__)) "\
  "%{!ansi: "\
    "-D_cdecl=__attribute__((__cdecl__)) "\
    "-D_stdcall=__attribute__((__stdcall__)) "\
    "-D_optlink=__attribute__((__optlink__))}"

/* Pass -s (as -Zstrip) to the assembler. */
#define ASM_SPEC "%{Zomf} %{s:-Zstrip}"

#define LINK_SPEC \
  "%{Zexe} %{Zstack*} %{Zmap*} %{Zomf:%{Zdll}}" \
  "%{!o*:-o %b%{Zdll:.dll}%{!Zdll:%{!Zexe:.exe}}}" \
  "%{static:%{Zcrtdll*:%e-static and -Zcrtdll are incompatible}}" \
  "%{Zdll:%{Zexe:%e-Zdll and -Zexe are incompatible}}" \
  "%{Zso:%{!Zsys:%e-Zso requires -Zsys}}" \
  "%{Zso:%{!Zdll:%e-Zso requires -Zdll}}" \
  "%{Zno-rte:%{!Zdll:%e-Zno-rte requires -Zdll}}" \
  "%{Zno-rte:%{!Zomf:%e-Zno-rte requires -Zomf}}" \
  "%{Zno-rte:%{Zsys:%e-Zno-rte and -Zsys are incompatible}}" \
  "%{Zno-rte:%{Zso:%e-Zno-rte and -Zso are incompatible}}" \
  "%{Zno-rte:%{Zcrtdll*:%e-Zcrtdll and -Zno-rte are incompatible}}" \
  "%{Zno-rte:%{Zmt:%e-Zmt and -Zno-rte are incompatible}}" \
  "%{Zsys:%{Zcrtdll*:%e-Zcrtdll and -Zsys are incompatible}}" \
  "%{Zsys:%{!Zomf:%e-Zsys requires -Zomf}}" \
  "%{Zsys:%{Zbsd-signals:%e-Zbsd-signals and -Zsys are incompatible}}" \
  "%{Zsys:%{Zsysv-signals:%e-Zsysv-signals and -Zsys are incompatible}}" \
  "%{Zbsd-signals:%{Zsysv-signals:%e-Zbsd-signals and -Zsysv-signals are incompatible}}" \
  "%{Zsmall-conv:%{Zcrtdll*:%e-Zsmall-conv and -Zcrtdll are incompatible}}"

#ifndef USE_CXX_RUNTIME
#define LIBGCC_SPEC "-lgcc%{Zcrtdll*:_dll}"
#else
#define LIBGCC_SPEC "-l" \
  "%{!Zcrtdll*:%{!fno-exceptions:gpp}}" \
  "%{!Zcrtdll*:%{fno-exceptions:gcc}}" \
  "%{Zcrtdll*:gcc_dll}"
#endif /* USE_CXX_RUNTIME */

/* "%{g*:-lg} " -- removed -- A.Z. */
#define LIB_SPEC \
  "%{p|pg:-lc_p} -lc " \
  "%{Zdll:%{Zso:-lc_dllso}%{Zno-rte:-lc_dllnrt} " \
         "%{!Zso:%{!Zno-rte:-lc_dllrt}}} " \
  "%{!Zdll:-lc_app%{pg:_p}} " \
  "%G %{p|pg:-lc_p} -lc " \
  "%{Zdll:%{!Zso:%{!Zno-rte:-lc_dllrt}}}"

#define STARTFILE_SPEC \
  "%{Zdll:dll0%O%{Zomf:bj}%s} " \
  "%{!Zdll:%{pg:gcrt0%O%{Zomf:bj}%s} " \
          "%{!pg:%{pn:gcrt0%O%{Zomf:bj}%s}%{!pn:%{p:mcrt0%O%{Zomf:bj}%s}%{!p:crt0%O%{Zomf:bj}%s}}} " \
          "%{Zsysv-signals:sigsysv%O%{Zomf:bj}%s} %{Zbsd-signals:sigbsd%O%{Zomf:bj}%s} " \
          "%{Zbin-files:binmode%O%{Zomf:bj}%s}%{Zsmall-conv:smallcnv%O%{Zomf:bj}%s}}"

#define LINK_COMMAND_SPEC \
  "%{!fsyntax-only:" \
  "%{!c:%{!M:%{!MM:%{!E:%{!S:%{!Zomf:ld}%{Zomf:emxomfld} " \
  "%l %X %{o*} %{A} %{d} %{e*} %{m} %{N} %{n} " \
  "%{r} %{s} %{t} %{u*} %{x} %{z} %{Z} " \
  "%{!A:%{!nostdlib:%{!nostartfiles:%S}}} " \
  "%{static:} %{L*} %D %{T*} %o " \
  "%{!nostdlib:%{!nodefaultlibs:-lc_alias %G "\
    "%{Zcrtdll=*:-lc_static -l%* -los2%{pg:_p} -lemx2} " \
    "%{!Zcrtdll*:%L %{Zno-rte:-los2%{pg:_p}} " \
      "%{!Zno-rte:%{Zsys:-lsys -los2} %{!Zsys:-lemx -los2%{pg:_p} -lemx2}}} " \
    "%{Zomf:-lend}}} " \
  "%{!A:%{!nostdlib:%{!nostartfiles:%E}}}}}}}}}"

/* IBM's linker LINK386 does not understand filenames with '+' (shame!) */
#define LIBSTDCXX "-lstdcxx"

/* Some additional system-dependent includes... */
#include <sys/emxload.h>
#include <sys/uflags.h>

/* EMX does not define strcasecmp but has stricmp instead */
#define strcasecmp stricmp

/* Do some OS/2-specific work upon initialization of all compilers */
#define COMPILER_HOST_INITIALIZATION                                           \
    /* Preload compiler if specified by GCCLOAD for faster subsequent runs */  \
    _emxload_env ("GCCLOAD");                                                  \
    /* Compilers don't fork (thanks God!) so we can use >32MB RAM */           \
    _uflags (_UF_SBRK_MODEL, _UF_SBRK_ARBITRARY);

/* emxomf does not understand stabs+, so for frontends we have to
   switch to standard stabs if -Zomf is used. We also do many other
   argv preprocessing here. */
#define GCC_DRIVER_HOST_INITIALIZATION                                         \
{                                                                              \
    int i, j, new_argc, max_argc, omf = 0, g_found = 0;                        \
    const char **new_argv;                                                     \
    _emxload_env ("GCCLOAD");                                                  \
    _envargs (&argc, (char ***)&argv, "GCCOPT");                               \
    _response (&argc, (char ***)&argv);                                        \
    _wildcard (&argc, (char ***)&argv);                                        \
    /* Copy argv into a new location and modify it while copying */            \
    new_argv = (const char **)malloc ((max_argc = argc) * sizeof (char *));    \
    new_argv [0] = argv [0];                                                   \
    for (i = 1, new_argc = 1; i < argc; i++)                                   \
    {                                                                          \
        int arg_count = 1;                                                     \
        const char *arg [4];                                                   \
        arg [0] = argv [i];                                                    \
        if (!strcmp (argv [i], "-Zomf"))                                       \
	  omf = 1;                                                             \
        else if (!strncmp (argv [i], "-g", 2)                                  \
              && ((argv [i][2] == 0) || ISDIGIT (argv [i][2])))                \
          g_found = new_argc;                                                  \
        else if (!strcmp (argv [i], "-Zmts"))                                  \
	  {                                                                    \
	    arg [0] = "-Zmt";                                                  \
	    arg [1] = "-static";                                               \
	    arg_count = 2;                                                     \
	  }                                                                    \
        else if (!strcmp (argv [i], "-Zmtd"))                                  \
	  {                                                                    \
	    arg [0] = "-Zmt";                                                  \
	    arg [1] = "-Zcrtdll=c_import";                                     \
	    arg_count = 2;                                                     \
	  }                                                                    \
        else if (!strcmp (argv [i], "-Zcrtdll"))                               \
	    arg [0] = "-Zcrtdll=c_import";                                     \
	else if (!strcmp (argv [i], "-Zlinker"))                               \
	  {                                                                    \
            if (i + 1 >= argc)                                                 \
	        fatal ("argument to `-Zlinker' is missing");                   \
	    arg [0] = "-Xlinker";                                              \
	    arg [1] = "-O";                                                    \
	    arg [2] = "-Xlinker";                                              \
	    arg [3] = argv [++i];                                              \
	    arg_count = 4;                                                     \
	  }                                                                    \
        if (new_argc + arg_count > max_argc)                                   \
            new_argv = (const char **)realloc (new_argv,		       \
                (new_argc + arg_count) * sizeof (char *));		       \
        for (j = 0; j < arg_count; j++)                                        \
            new_argv [new_argc++] = arg [j];                                   \
    }                                                                          \
    argv = new_argv; argc = new_argc;                                          \
    if (g_found && omf)                                                        \
    {                                                                          \
        char *temp = (char *)malloc (strlen (argv [g_found]) + 5 + 1);         \
        strcpy (temp, "-gstabs");                                              \
        strcat (temp, &argv [g_found][2]);                                     \
        ((char **)argv) [g_found] = temp;                                      \
    }                                                                          \
    INITIALIZE_PREFIXES							       \
}

/* Get perform_* macros to build libgcc.a.  */
#include "i386/perform.h"

/* The following will be defined only when compiling libgcc2
 * to avoid including large .h files when they are not needed.
 * Including them is neccessary to avoid 'implicit declatation'
 * warnings, and also for __write.
 */
#ifdef IN_LIBGCC2
/* define "write" and some other external functions for libgcc */
#define write __write

#include <malloc.h>
#include <alloca.h>
#include <time.h>
#include <string.h>

extern void __ctordtorInit (void);
extern void __ctordtorTerm (void);

/* Invoke static object constructors/destructors */
/* Also register/deregister frame unwind information */
#define DO_GLOBAL_CTORS_BODY __ctordtorInit ()
#define DO_GLOBAL_DTORS_BODY __ctordtorTerm ()

/* We call register_frame_table ourselves */
#define DONT_AUTOREGISTER_FRAME_INFO

#endif
